#!/usr/bin/env python3
"""Refresh the BUILD file of SPDX license_kinds with new ones from licenses.json.

Usage:
  wget https://github.com/spdx/license-list-data/raw/master/json/licenses.json
  LC_ALL="en_US.UTF-8" admin/refresh_spdx/add_licenses.py
  git diff
  git commit
"""

import json
import os
from string import Template
import re
import sys

def load_json(path: str):
  """Loads a JSON file and returns the data.

  Args:
    path: (str) a file path.
  Returns:
    (dict) the parsed data.
  """
  with open(path, 'r') as fp:
    ret = json.load(fp)
  return ret


def parse_build_file(text: str):
  """Parse a BUILD file of license declaration.

  Parses a build file and returns all the text that
  is not licenese_kind declarations and a map of
  license kind names to the text which declared them

  license_kind(
    name = "0BSD",
    conditions = [],
    url = "https://spdx.org/licenses/0BSD.html",
  )

  Returns:
    preamble: str 
    targets: map<str, str>
  """

  target_start = re.compile(r'^([a-zA-Z0-9_]+)\(')  # ) to fix autoindent
  name_match = re.compile(r'^ *name *= *"([^"]*)"')

  targets = {}
  preamble = []
  collecting_rule = None

  for line in text.split('\n'):
    if collecting_rule:
      collecting_rule.append(line)
      if line.startswith(')'):
        assert cur_name
        targets[cur_name] = '\n'.join(collecting_rule)
        # print(cur_name, "=====", targets[cur_name])
        collecting_rule = None
        cur_name = None
      else:
        m = name_match.match(line)
        if m:
          cur_name = m.group(1)
      continue

    # not collecting a rule
    m = target_start.match(line)
    if m:
      cur_rule = m.group(1)
      if cur_rule == 'license_kind':
        collecting_rule = [line]
        continue

    if line or preamble[-1]:
      preamble.append(line)

  return '\n'.join(preamble), targets


# Sample JSON formate license declaration.
"""
{
  "licenseListVersion": "3.8-57-gca4f142",
  "licenses": [
    {
      "reference": "./0BSD.html",
      "isDeprecatedLicenseId": false,
      "detailsUrl": "http://spdx.org/licenses/0BSD.json",
      "referenceNumber": "232",
      "name": "BSD Zero Clause License",
      "licenseId": "0BSD",
      "seeAlso": [
        "http://landley.net/toybox/license.html"
      ],
      "isOsiApproved": true
    },
"""


def insert_new(already_declared, licenses):
  template = Template((
      'license_kind(\n'
      '    name = "${licenseId}",\n'
      '    conditions = [],\n'
      '    url = "https://spdx.org/licenses/${licenseId}.html",\n'
      ')'))

  for license in licenses:
    id = license['licenseId']
    old = already_declared.get(id)
    if not old:
      print('Adding:', id)
      already_declared[id] = template.substitute(license)


def update_spdx_build():
  """Update //licenses/spdx/BUILD with new license kinds."""

  license_json = load_json('licenses.json')

  # Find workspace root
  build_path = 'licenses/spdx/BUILD'
  found_spdx_build = False
  for i in range(5):  # Simple regression failure limiter
    if os.access('WORKSPACE', os.R_OK) and os.access(build_path, os.R_OK):
      found_spdx_build = True
      break
    os.chdir('..')
  if not found_spdx_build:
    print('Could not find', build_path)
    return 1

  with open(build_path, 'r') as fp:
    current_file_content = fp.read()

  preamble, kinds = parse_build_file(current_file_content)
  insert_new(kinds, license_json['licenses'])
  with open(build_path, 'w') as fp:
    fp.write(preamble)
    for license in sorted(kinds.keys(), key=lambda x: x.lower()):
      fp.write('\n')
      fp.write(kinds[license])
      fp.write('\n')


def main():
  lc_all = os.environ.get('LC_ALL')
  if lc_all != 'en_US.UTF-8':
    print('Your environment settings will reorder the file badly.')
    print('Please rerun as:')
    print('  LC_ALL="en_US.UTF-8"', ' '.join(sys.argv))
    sys.exit(1)

  update_spdx_build()


if __name__ == '__main__':
  main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import unittest
from tests import license_test_utils


class AnAppLicensesTest(unittest.TestCase):

  def test_has_expected_licenses(self):
    package_base = license_test_utils.LICENSE_PACKAGE_BASE
    licenses_info = license_test_utils.load_licenses_info(
        os.path.join(os.path.dirname(__file__), "an_app_licenses.json"))
    licenses_info = license_test_utils.filter_dependencies(
        licenses_info,
        target_filter=lambda targ: targ.startswith(package_base),
        licenses_filter=lambda lic: lic.startswith(package_base))

    expected = {
        "/tests/thrdparty:new_style_lib": [
            "/tests/thrdparty:license",
        ],
    }
    license_test_utils.check_licenses_of_dependencies(
        self, licenses_info, expected)


if __name__ == "__main__":
  unittest.main()



"""Utilities for writing tests of license rules."""

import codecs
import json

# This is extracted out to make it easier to keep test equivalence between
# the OSS version and Google.
LICENSE_PACKAGE_BASE = "/"


def load_licenses_info(info_path):
  """Loads the licenses_info() JSON format."""
  with codecs.open(info_path, encoding="utf-8") as licenses_file:
    return json.loads(licenses_file.read())


def filter_dependencies(licenses_info, target_filter=None,
                        licenses_filter=None):
  """Filters licenses_info to only include dependencies of interest.

  Args:
    licenses_info: (dict) licenses info.
    target_filter: (function): function which returns true if we should include
        the target.
    licenses_filter: (function): function which returns true if we should
        include the license.
  Returns:
    (dict) a valid licenses_info dict.
  """
  top_target = licenses_info[0]
  new_top_target = dict(top_target)
  new_deps = []
  for dep in top_target["dependencies"]:
    target_name = dep["target_under_license"]
    if target_filter and not target_filter(target_name):
      continue
    licenses = dep["licenses"]
    if licenses_filter:
      licenses = [lic for lic in licenses if licenses_filter(lic)]
    new_deps.append({
        "target_under_license": target_name,
        "licenses": licenses})
  new_top_target["dependencies"] = new_deps
  return [new_top_target]


def check_licenses_of_dependencies(test_case, licenses_info, expected,
                                   path_prefix=LICENSE_PACKAGE_BASE):
  """Checks that licenses_info contains an expected set of licenses.

  Args:
    test_case: (TestCase) the test.
    licenses_info: (dict) licenses info.
    expected: (dict) map of target name suffixes to the licenses they are under.
    path_prefix: (str) prefix to prepend to targets and licenses in expected.
       This turns the relative target names to absolute ones.
  """

  # Turn the list of deps into a dict by target for easier comparison.
  deps_to_licenses = {
      x["target_under_license"].lstrip("@"): set(l.strip("@") for l in x["licenses"])
      for x in licenses_info[0]["dependencies"]}

  target_names = ",".join(deps_to_licenses.keys())
  # This is n**2, but N is typically < 3 or we are doing this wrong.
  for want_target, want_licenses in expected.items():
    found_target = False
    for got_target, got_licenses in deps_to_licenses.items():
      if got_target.endswith(want_target):
        found_target = True
        test_case.assertEqual(len(want_licenses), len(got_licenses))
        found_license = False
        for want_l in want_licenses:
          for got_l in got_licenses:
            if got_l.endswith(want_l):
              found_license = True
              break
        test_case.assertTrue(
            found_license,
            msg="license (%s) not a suffix in %s" % (want_l, got_licenses))
        break
    test_case.assertTrue(
        found_target,
        msg="target (%s) not a suffix in [%s]" % (want_target, target_names))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A trivial tool to turn a string into a C++ constant.

This is not meant to be useful. It is only to provide an example of a tool that
generates code.
"""

import sys


def main(argv):
  if len(argv) != 4:
    raise Exception('usage: constant_generator out_file var_name text')
  with open(argv[1], 'w') as out:
    out.write('const char* %s = "%s";\n' % (argv[2], argv[3]))


if __name__ == '__main__':
  main(sys.argv)



"""Test that we see the expected licenses for some generated code."""

import codecs
import os

import unittest
from tests import license_test_utils

class VerifyLicensesTest(unittest.TestCase):

  def test_generater_license(self):
    licenses_info = license_test_utils.load_licenses_info(
        os.path.join(os.path.dirname(__file__), "generator_licenses.json"))

    expected = {
        "examples/vndor/constant_gen:constant_generator": [
            "examples/vndor/constant_gen:license"
        ],
    }
    license_test_utils.check_licenses_of_dependencies(
        self, licenses_info, expected)

  def test_generated_code_license(self):
    licenses_info = license_test_utils.load_licenses_info(
        os.path.join(os.path.dirname(__file__), "generated_code_licenses.json"))

    expected = {
        "examples/vndor/constant_gen:libhello": [
            "/constant_gen:license_for_emitted_code",
        ],
        "examples/vndor/constant_gen:libhello_src_": [
            "/constant_gen:license_for_emitted_code",
        ],
    }
    license_test_utils.check_licenses_of_dependencies(
        self, licenses_info, expected)

if __name__ == "__main__":
  unittest.main()




"""Tests for google3.tools.build_defs.license.tests.hello_licenses."""

import codecs
import os

import unittest
from tests import license_test_utils


class HelloLicensesTest(unittest.TestCase):

  def test_has_expected_licenses(self):
    licenses_info = license_test_utils.load_licenses_info(
        os.path.join(os.path.dirname(__file__), "hello_licenses.json"))

    expected = {
        "/tests:hello": [
            "/tests:license",
        ],
        "/tests:c_bar": [
            "/tests:license",
            "/tests:license_for_extra_feature",
        ],
    }
    license_test_utils.check_licenses_of_dependencies(
        self, licenses_info, expected)

  def test_has_expected_copyrights(self):
    copyrights_file = os.path.join(os.path.dirname(__file__),
                                   "hello_cc_copyrights.txt")
    with codecs.open(copyrights_file, encoding="utf-8") as inp:
      copyrights = inp.read().split('\n')
      self.assertIn(
          "package(A test case package/0.0.4), copyright(Copyright © 2019 Uncle Toasty)",
          copyrights)
      self.assertIn(
          "package(A test case package), copyright()",
          copyrights)


if __name__ == "__main__":
  unittest.main()



#!/usr/bin/env python3
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Proof of concept license checker.

This is only a demonstration. It will be replaced with other tools.
"""

import argparse
import codecs
import datetime
import json
import os


TOOL = 'https//github.com/bazelbuild/rules_license/tools:write_sbom'

def _load_package_data(package_info):
  with codecs.open(package_info, encoding='utf-8') as inp:
    return json.loads(inp.read())

def _write_sbom_header(out, package):
  header = [
    'SPDXVersion: SPDX-2.2',
    'DataLicense: CC0-1.0',
    'SPDXID: SPDXRef-DOCUMENT',
    'DocumentName: %s' % package,
    # TBD
    # 'DocumentNamespace: https://swinslow.net/spdx-examples/example1/hello-v3
    'Creator: Person: %s' % os.getlogin(),
    'Creator: Tool: %s' % TOOL,
    datetime.datetime.utcnow().strftime('Created: %Y-%m-%d-%H:%M:%SZ'),
    '',
    '##### Package: %s' % package,
  ]
  out.write('\n'.join(header))



def _write_sbom(out, packages):
  """Produce a basic SBOM

  Args:
    out: file object to write to
    packages: package metadata. A big blob of JSON.
  """
  for p in packages:
    name = p.get('package_name') or '<unknown>'
    out.write('\n')
    out.write('SPDXID: "%s"\n' % name)
    out.write('  name: "%s"\n' % name)
    if p.get('package_version'):
      out.write('  versionInfo: "%s"\n' % p['package_version'])
    # IGNORE_COPYRIGHT: Not a copyright notice. It is a variable holding one.
    cn = p.get('copyright_notice')
    if cn:
      out.write('  copyrightText: "%s"\n' % cn)
    kinds = p.get('license_kinds')
    if kinds:
      out.write('  licenseDeclared: "%s"\n' %
                ','.join([k['name'] for k in kinds]))
    url = p.get('package_url')
    if url:
      out.write('  downloadLocation: %s\n' % url)


def main():
  parser = argparse.ArgumentParser(
      description='Demonstraton license compliance checker')

  parser.add_argument('--licenses_info',
                      help='path to JSON file containing all license data')
  parser.add_argument('--out', default='sbom.out', help='SBOM output')
  args = parser.parse_args()

  license_data = _load_package_data(args.licenses_info)
  target = license_data[0]  # we assume only one target for the demo

  top_level_target = target['top_level_target']
  dependencies = target['dependencies']
  # It's not really packages, but this is close proxy for now
  licenses = target['licenses']
  package_infos = target['packages']

  # These are similar dicts, so merge them by package. This is not
  # strictly true, as different licenese can appear in the same
  # package, but it is good enough for demonstrating the sbom.

  all = {x['bazel_package']: x for x in licenses}
  for pi in package_infos:
    p = all.get(pi['bazel_package'])
    if p:
      p.update(pi)
    else:
      all[pi['bazel_package']] = pi

  err = 0
  with codecs.open(args.out, mode='w', encoding='utf-8') as out:
    _write_sbom_header(out, package=top_level_target)
    _write_sbom(out, all.values())
  return err


if __name__ == '__main__':
  main()



#!/usr/bin/env python3
# Copyright 2022 The Bazel Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""merge stardoc output into a single page.

- concatenates files
- corrects things that stardoc botches
"""

import re
import sys
import typing


# I think stardoc changed the format of the id strings. Sigh.
ID_RE = re.compile(r'<a id="(.*)">')
ID_OLD_RE = re.compile(r'<a id="#(.*)">')
WRAPS_RE = re.compile(r'@wraps\((.*)\)')
SINCE_RE = re.compile(r'@since\(([^)]*)\)')
CENTER_RE = re.compile(r'<p align="center">([^<]*)</p>')


def merge_file(file: str, out, wrapper_map:typing.Dict[str, str]) -> None:
  with open(file, 'r') as inp:
    content = inp.read()
    m = ID_RE.search(content)
    if not m:
      m = ID_OLD_RE.search(content)
    this_rule = m.group(1) if m else None
    m = WRAPS_RE.search(content)
    if m:
      # I wrap something, so don't emit me.
      wrapper_map[m.group(1)] = this_rule
      return
    # If something wraps me, rewrite myself with the wrapper name.
    if this_rule in wrapper_map:
      content = content.replace(this_rule, wrapper_map[this_rule])
    merge_text(content, out)


def merge_text(text: str, out) -> None:
  """Merge a block of text into an output stream.

  Args:
    text: block of text produced by Starroc.
    out: an output file stream.
  """
  for line in text.split('\n'):
    line = SINCE_RE.sub(r'<div class="since"><i>Since \1</i></div>', line)

    if line.startswith('| :'):
      line = fix_stardoc_table_align(line)
    # Compensate for https://github.com/bazelbuild/stardoc/issues/118.
    # Convert escaped HTML <li> back to raw text
    line = line.replace('&lt;li&gt;', '<li>')
    line = CENTER_RE.sub(r'\1', line)
    _ = out.write(line)
    _ = out.write('\n')


def fix_stardoc_table_align(line: str) -> str:
  """Change centered descriptions to left justified."""
  if line.startswith('| :-------------: | :-------------: '):
    return '| :------------ | :--------------- | :---------: | :---------: | :----------- |'
  return line


def main(argv: typing.Sequence[str]) -> None:
  wrapper_map = {}
  for file in argv[1:]: 
    merge_file(file, sys.stdout, wrapper_map)


if __name__ == '__main__':
  main(sys.argv)



#!/usr/bin/env python3
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Proof of concept license checker.

This is only a demonstration. It will be replaced with other tools.
"""

import argparse
import codecs
import json

# Conditions allowed for all applications
_ALWAYS_ALLOWED_CONDITIONS = frozenset(['notice', 'permissive', 'unencumbered'])


def _load_license_data(licenses_info):
  with codecs.open(licenses_info, encoding='utf-8') as licenses_file:
    return json.loads(licenses_file.read())


def unique_licenses(licenses):
  for target in licenses:
    for lic in target.get('licenses') or []:
      yield lic

def _do_report(out, licenses):
  """Produce a report showing the set of licenses being used.

  Args:
    out: file object to write to
    licenses: list of LicenseInfo objects

  Returns:
    0 for no restricted licenses.
  """

  for target in unique_licenses(licenses):
    for lic in target.get('licenses') or []:
      print("lic:", lic)
      rule = lic['rule']
      for kind in lic['license_kinds']:
        out.write('= %s\n  kind: %s\n' % (rule, kind['target']))
        out.write('  conditions: %s\n' % kind['conditions'])


def _check_conditions(out, licenses, allowed_conditions):
  """Check that the application does not use any disallowed licenses.

  Args:
    out: file object to write to
    licenses: list of LicenseInfo objects
    allowed_conditions: list of allowed condition names

  Returns:
    0 for no licenses from outside allowed_conditions.
  """
  err = 0
  for lic in licenses:  # using strange name lic because license is built-in
    rule = lic['rule']
    for kind in lic['license_kinds']:
      disallowed = []
      for condition in kind['conditions']:
        if condition not in allowed_conditions:
          disallowed.append(condition)
      if disallowed:
        out.write('ERROR: %s\n' % rule)
        out.write('   kind: %s\n' % kind['target'])
        out.write('   conditions: %s\n' % kind['conditions'])
        out.write('   disallowed condition: %s\n' % ','.join(disallowed))
        err += 1
  return err


def _do_copyright_notices(out, licenses):
  for l in licenses:
    name = l.get('package_name') or '<unknown>'
    if l.get('package_version'):
      name =  name + "/" + l['package_version']
    # IGNORE_COPYRIGHT: Not a copyright notice. It is a variable holding one.
    out.write('package(%s), copyright(%s)\n' % (name, l['copyright_notice']))


def _do_licenses(out, licenses):
  for lic in unique_licenses(licenses):
    path = lic['license_text']
    with codecs.open(path, encoding='utf-8') as license_file:
      out.write('= %s\n' % path)
      out.write(license_file.read())


def main():
  parser = argparse.ArgumentParser(
      description='Demonstraton license compliance checker')

  parser.add_argument('--licenses_info',
                      help='path to JSON file containing all license data')
  parser.add_argument('--report', default='report', help='Summary report')
  parser.add_argument('--copyright_notices', 
                      help='output file of all copyright notices')
  parser.add_argument('--license_texts', help='output file of all license files')
  parser.add_argument('--check_conditions', action='store_true',
                      help='check that the dep only includes allowed license conditions')
  args = parser.parse_args()

  license_data = _load_license_data(args.licenses_info)
  target = license_data[0]  # we assume only one target for the demo

  top_level_target = target['top_level_target']
  dependencies = target['dependencies']
  licenses = target['licenses']

  err = 0
  with codecs.open(args.report, mode='w', encoding='utf-8') as rpt:
    _do_report(rpt, licenses)
    if args.check_conditions:
      # TODO(aiuto): Read conditions from a file of allowed conditions for
      # a specified application deployment environment.
      err = _check_conditions(rpt, licenses, _ALWAYS_ALLOWED_CONDITIONS)
  if args.copyright_notices:
    with codecs.open(
        args.copyright_notices, mode='w', encoding='utf-8') as out:
      _do_copyright_notices(out, licenses)
  if args.license_texts:
    with codecs.open(args.license_texts, mode='w', encoding='utf-8') as out:
      _do_licenses(out, licenses)
  return err


if __name__ == '__main__':
  main()
